@@ -10,10 +10,11 @@
 # http://opensource.org/licenses/mit-license.php
 # =================================================================
 
-__version__ = "0.4.0"
+__version__ = "0.4.1"
 
 import argparse
 from argparse import ArgumentParser, Action, Namespace
+from collections import namedtuple
 from contextlib import contextmanager
 from getpass import getpass
 import glob
@@ -45,12 +46,53 @@ LOG_FORMAT: str = '%(asctime)s |  %(levelname)-7s | %(message)s (%(filename)s L%
 
 INTERVAL: int = 1
 
+
+def parse_date_str_int(date_s: str) -> tuple[str, str, str]:
+    match: re.Match = re.search("(\d+)\s*年\s*(\d+)\s*月\s*(\d+)\s*日", date_s)
+    if match:
+        year_i: int = int(match.group(1))
+        month_i: int = int(match.group(2))
+        day_i: int = int(match.group(3))
+        return year_i, month_i, day_i
+    raise ValueError("Could not parse date_s '%s'", date_s)
+
+
 ################################################################################
 # Utilities for Selenium
 
 
+def get_driver_and_wait(download_dir: str = None, browser: str = None) -> tuple[WebDriver, WebDriverWait, int]:
+
+    timeout_sec: int = 60
+
+    # download_dir: str = "/Users/nogayama1/Downloads"
+    if not download_dir:
+        # download_dir = os.path.join(os.path.expanduser("~"), "Downloads")
+        download_dir = os.path.join(os.getcwd(), "Downloads")
+
+    if not browser:
+        browser = "firefox"
+
+    driver: WebDriver = None
+    if browser == "chrome":
+        chrome_options = webdriver.ChromeOptions()
+        prefs = {'download.default_directory': download_dir}
+        chrome_options.add_experimental_option('prefs', prefs)
+        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), chrome_options=chrome_options)
+    elif browser == "firefox":
+        options = FirefoxOptions()
+        options.set_preference("browser.download.folderList", 2)
+        options.set_preference("browser.download.dir", download_dir)
+        driver = webdriver.Firefox(service=FirefoxService(GeckoDriverManager().install()), options=options)
+    else:
+        raise ValueError(f"browser type '{browser}' is not supported.")
+    driver.implicitly_wait(timeout_sec)
+    wait: WebDriverWait = WebDriverWait(driver, timeout_sec)
+    return driver, wait, timeout_sec
+
+
 @contextmanager
-def inspect_mode(driver, original_timeout, inspect_sec=1):
+def inspect_mode(driver: WebDriver, original_timeout: int, inspect_sec: int = 1):
     driver.implicitly_wait(inspect_sec)
     wait2 = WebDriverWait(driver, inspect_sec)
     try:
@@ -63,7 +105,7 @@ def inspect_mode(driver, original_timeout, inspect_sec=1):
         driver.implicitly_wait(original_timeout)
 
 
-def download_is_completed(download_dir, last_newest_file):
+def download_is_completed(download_dir: str, last_newest_file: str = None):
 
     def _f(driver):
         filepaths: list[str] = glob.glob(os.path.join(download_dir, "*"))
@@ -96,13 +138,25 @@ def download_is_completed(download_dir, last_newest_file):
     return _f
 
 
+DownloadResult = namedtuple('DownloadResult', 'downloaded_filepath')
+
+
+@contextmanager
+def safe_download(driver: WebDriver, wait: WebDriverWait, download_dir: str, target_filepath_prefix: str):
+
+    filepaths: list[str] = glob.glob(os.path.join(download_dir, "*"))
+    last_newest_file = max(filepaths, key=lambda fp: os.path.getmtime(fp)) if filepaths else None
+
+    download_result = DownloadResult()
+    yield download_result
+
+    _LOGGER.debug("Waiting until the download is completed")
+    download_result.downloaded_filepath = wait.until(download_is_completed(download_dir, last_newest_file))
+
+
 ################################################################################
 # Utilities for Wellnote
-
-
-def download_album(start_year: int = 2009, start_month: int = 1, \
-                   end_year: int = 2023, end_month: int = 12,  \
-                   download_dir: str = None, browser: str = None) -> int:
+def get_email_and_password() -> tuple[str, str]:
 
     email: str = None
     password: str = None
@@ -125,31 +179,11 @@ def download_album(start_year: int = 2009, start_month: int = 1, \
     if no_environ_vars:
         _LOGGER.warning("You can omit username and password prompt by declaring environmental variables WELLNOTE_EMAIL and WELLNOTE_PASSWORD.")
 
-    # download_dir: str = "/Users/nogayama1/Downloads"
-    if not download_dir:
-        # download_dir = os.path.join(os.path.expanduser("~"), "Downloads")
-        download_dir = os.path.join(os.getcwd(), "Downloads")
+    return email, password
 
-    if not browser:
-        browser = "firefox"
-    timeout_sec: int = 60
-
-    driver: WebDriver = None
-    if browser == "chrome":
-        chrome_options = webdriver.ChromeOptions()
-        prefs = {'download.default_directory': download_dir}
-        chrome_options.add_experimental_option('prefs', prefs)
-        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), chrome_options=chrome_options)
-    elif browser == "firefox":
-        options = FirefoxOptions()
-        options.set_preference("browser.download.folderList", 2)
-        options.set_preference("browser.download.dir", download_dir)
-        driver = webdriver.Firefox(service=FirefoxService(GeckoDriverManager().install()), options=options)
-    else:
-        raise ValueError(f"browser type '{browser}' is not supported.")
-    driver.implicitly_wait(timeout_sec)
-    wait = WebDriverWait(driver, timeout_sec)
 
+@contextmanager
+def wellnote(driver: WebDriver, wait: WebDriverWait, email: str, password: str):
     try:
         _LOGGER.info("Geting wellnote.jp")
         driver.get("https://wellnote.jp/")
@@ -181,163 +215,182 @@ def download_album(start_year: int = 2009, start_month: int = 1, \
             password_form.send_keys(Keys.ENTER)
             time.sleep(INTERVAL)
 
-            if True:
-                wait.until(EC.staleness_of(password_form))
+            wait.until(EC.staleness_of(password_form))
 
-                _LOGGER.debug("Waiting until a clickable albums button is available")
-                # <a class="sc-jWWnA hivVBT" href="/albums">
-                album_button: WebElement = wait.until(EC.element_to_be_clickable([By.XPATH, "//a[@href='/albums']"]))
+            yield
 
-                _LOGGER.info("Clicking the album button")
-                album_button.click()
-                time.sleep(INTERVAL)
-
-                if True:
-                    ## Go to start year
-                    year: int = 9999
-                    while year >= 2009:  # We can't go back to the years before wellnote's inception
-
-                        _LOGGER.debug("Waiting until a year text is available")
-                        # year = wait.until(EC.visibility_of_element_located([By.XPATH, "//div[contains(text(), '年')]"])) # dont work
-                        year_text: str = wait.until(EC.visibility_of_element_located([By.XPATH, "//div[@class='sc-bOtlzW fgPidp']"]))
-                        year = int(year_text.text.replace("年", ""))
-
-                        if year == start_year:
-                            break
-
-                        move_previous_year_button = None
-                        with inspect_mode(driver, timeout_sec) as wait2:
-                            _LOGGER.debug("Waiting until a clickable next button is available")
-                            move_previous_year_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-jFkwbb ruuVe']"]))
-                        if not move_previous_year_button:
-                            _LOGGER.info("Breaking this year because previous button is not found")
-                            start_month = 1
-                            break
-
-                        _LOGGER.info("Moving previous year of %s ", year)
-                        move_previous_year_button.click()
-                        time.sleep(INTERVAL)
-
-                    ## Iterate over years
-                    while year <= end_year:
-
-                        _LOGGER.info("Starting year %s", year)
-
-                        end_month_of_this_year = 12 if year != end_year else end_month
-                        month: int
-                        for month in range(start_month, end_month_of_this_year + 1):
-
-                            _LOGGER.debug("Waiting until a clickable %s-th month button is available", month)
-                            month_button: WebElement = wait.until(EC.element_to_be_clickable([By.XPATH, f"//li[text()='{month}']"]))
-                            if "jAQwgG" in month_button.get_attribute("class"):
-                                # selected
-                                _LOGGER.info("Already at month %s", month)
-                            else:
-                                # if month_button.is_displayed() and month_button.is_enabled(): # dont work
-                                if "RJZN" in month_button.get_attribute("class"):
-                                    _LOGGER.info("Moving %s-th month", month)
-                                    month_button.click()
-                                    time.sleep(INTERVAL)
-                                else:
-                                    _LOGGER.info("Found month %s does not have data", month)
-                                    continue
-
-                            _LOGGER.debug("Waiting until a clickable upper left grid item is available")
-                            first_grid_item: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "virtuoso-grid-item"]))
-
-                            _LOGGER.info("Clicking the upper left grid item")
-                            first_grid_item.click()
-                            time.sleep(INTERVAL)
-
-                            idx: int = 0
-                            last_date_s = None
-                            while True:
-
-                                _LOGGER.debug("Waiting until a visible date text is available")
-                                date_elem: WebElement = wait.until(EC.visibility_of_element_located([By.CLASS_NAME, "sc-dWbSDx"]))
-                                date_s: str = date_elem.text
-                                _LOGGER.info("Found date %s", date_s)
-
-                                if date_s != last_date_s:
-                                    idx = 0
-                                    last_date_s = date_s
-
-                                match: re.Match = re.search("(\d+)\s*年\s*(\d+)\s*月\s*(\d+)\s*日", date_s)
-                                year_i: int = int(match.group(1))
-                                month_i: int = int(match.group(2))
-                                day_i: int = int(match.group(3))
-
-                                tareget_basename = f"wellnote_{year_i:04}-{month_i:02}-{day_i:02}_{idx:03}"
-                                target_dir = os.path.join(download_dir, "wellnote", f"{year_i:04}")
-                                target_filepath_woe = os.path.join(target_dir, tareget_basename)
-
-                                # if os.path.exists(target_filepath):
-                                if glob.glob(target_filepath_woe + ".*"):
-                                    _LOGGER.warning("Skipping    %s because it exists", target_filepath_woe.replace(os.getcwd(), "."))
-                                else:
-                                    _LOGGER.warning("Downloading %s", target_filepath_woe.replace(os.getcwd(), "."))
-
-                                    filepaths: list[str] = glob.glob(os.path.join(download_dir, "*"))
-                                    last_newest_file = max(filepaths, key=lambda fp: os.path.getmtime(fp)) if filepaths else None
-
-                                    _LOGGER.debug("Waiting until a clickable vdots button is available")
-                                    vdots_button: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "sc-hCwLRM"]))
-                                    _LOGGER.info("Clicking vdots button")
-                                    vdots_button.click()
-
-                                    _LOGGER.debug("Waiting until a clickable download button is available")
-                                    download_button: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "sc-hLVXRe"]))
-                                    _LOGGER.info("Clicking download button")
-                                    download_button.click()
-
-                                    time.sleep(INTERVAL)
-
-                                    _LOGGER.debug("Waiting until the download is completed")
-                                    downloaded_filepath = wait.until(download_is_completed(download_dir, last_newest_file))
-
-                                    extension: str = downloaded_filepath.split(".")[-1]
-                                    target_filepath = target_filepath_woe + "." + extension
-
-                                    os.makedirs(os.path.join(target_dir), exist_ok=True)
-                                    shutil.move(downloaded_filepath, target_filepath)
-
-                                swiper_button_next = None
-                                with inspect_mode(driver, timeout_sec) as wait2:
-                                    _LOGGER.debug("Waiting until a clickable next button is available")
-                                    swiper_button_next: WebElement = wait2.until(EC.element_to_be_clickable([By.CLASS_NAME, "swiper-button-next"]))
-                                if not swiper_button_next \
-                                or "swiper-button-disabled" in swiper_button_next.get_attribute("class"):
-                                    _LOGGER.info("Breaking this month because next button is not found")
-                                    break
-
-                                _LOGGER.info("Clicking the swiper_button_next")
-                                swiper_button_next.click()
-                                # time.sleep(INTERVAL)
-
-                                idx += 1
-
-                            _LOGGER.debug("Waiting until a clickable html body is available")
-                            close_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-edERGn jYcwJ']"]))
-
-                            _LOGGER.info("Closing the preview window")
-                            close_button.click()
-
-                        move_next_year_button = None
-                        with inspect_mode(driver, timeout_sec) as wait2:
-                            _LOGGER.debug("Waiting until a clickable next year button is available")
-                            move_next_year_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-jFkwbb lbXnTj']"]))
-                        if not move_next_year_button:
-                            _LOGGER.info("Breaking this year because next year button is not found")
-                            break
-
-                        _LOGGER.info("Moving the next year of %s ", year)
-                        move_next_year_button.click()
-                        time.sleep(INTERVAL)
-
-                        year += 1
-                        start_month = 1
     finally:
         driver.quit()
+
+
+@contextmanager
+def album_tab(driver: WebDriver, wait: WebDriverWait):
+
+    _LOGGER.debug("Waiting until a clickable albums button is available")
+    # <a class="sc-jWWnA hivVBT" href="/albums">
+    album_button: WebElement = wait.until(EC.element_to_be_clickable([By.XPATH, "//a[@href='/albums']"]))
+
+    _LOGGER.info("Clicking the album button")
+    album_button.click()
+    time.sleep(INTERVAL)
+
+    yield
+
+def download_album(start_year: int = 2009, start_month: int = 1, \
+                   end_year: int = 2023, end_month: int = 12, \
+                   download_dir: str = None, browser: str = None) -> int:
+    _LOGGER.info("Invoking download_album(start_year='%s', start_month='%s', end_year='%s', end_month='%s', download_dir='%s', browser='%s')")
+
+    email: str
+    password: str
+    email, password = get_email_and_password()
+
+    driver: WebDriver
+    wait: WebDriverWait
+    timeout_sec: int
+    driver, wait, timeout_sec = get_driver_and_wait(download_dir, browser)
+
+    with wellnote(driver, wait, email, password):
+
+        with album_tab(driver, wait):
+
+            ## Go to start year
+            year: int = 9999
+            while year >= 2009:  # We can't go back to the years before wellnote's inception
+
+                _LOGGER.debug("Waiting until a year text is available")
+                # year = wait.until(EC.visibility_of_element_located([By.XPATH, "//div[contains(text(), '年')]"])) # dont work
+                year_text: str = wait.until(EC.visibility_of_element_located([By.XPATH, "//div[@class='sc-bOtlzW fgPidp']"]))
+                year = int(year_text.text.replace("年", ""))
+
+                if year == start_year:
+                    break
+
+                move_previous_year_button = None
+                with inspect_mode(driver, timeout_sec) as wait2:
+                    _LOGGER.debug("Waiting until a clickable next button is available")
+                    move_previous_year_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-jFkwbb ruuVe']"]))
+                if not move_previous_year_button:
+                    _LOGGER.info("Breaking this year because previous button is not found")
+                    start_month = 1
+                    break
+
+                _LOGGER.info("Moving previous year of %s ", year)
+                move_previous_year_button.click()
+                time.sleep(INTERVAL)
+
+            ## Iterate over years
+            while year <= end_year:
+
+                _LOGGER.info("Starting year %s", year)
+
+                end_month_of_this_year = 12 if year != end_year else end_month
+                month: int
+                for month in range(start_month, end_month_of_this_year + 1):
+
+                    _LOGGER.debug("Waiting until a clickable %s-th month button is available", month)
+                    month_button: WebElement = wait.until(EC.element_to_be_clickable([By.XPATH, f"//li[text()='{month}']"]))
+                    if "jAQwgG" in month_button.get_attribute("class"):
+                        # selected
+                        _LOGGER.info("Already at month %s", month)
+                    else:
+                        # if month_button.is_displayed() and month_button.is_enabled(): # dont work
+                        if "RJZN" in month_button.get_attribute("class"):
+                            _LOGGER.info("Moving %s-th month", month)
+                            month_button.click()
+                            time.sleep(INTERVAL)
+                        else:
+                            _LOGGER.info("Found month %s does not have data", month)
+                            continue
+
+                    _LOGGER.debug("Waiting until a clickable upper left grid item is available")
+                    first_grid_item: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "virtuoso-grid-item"]))
+
+                    _LOGGER.info("Clicking the upper left grid item")
+                    first_grid_item.click()
+                    time.sleep(INTERVAL)
+
+                    idx: int = 0
+                    last_date_s = None
+                    while True:
+
+                        _LOGGER.debug("Waiting until a visible date text is available")
+                        date_elem: WebElement = wait.until(EC.visibility_of_element_located([By.CLASS_NAME, "sc-dWbSDx"]))
+                        date_s: str = date_elem.text
+                        _LOGGER.info("Found date %s", date_s)
+
+                        if date_s != last_date_s:
+                            idx = 0
+                            last_date_s = date_s
+
+                        year_i, month_i, day_i = parse_date_str_int(date_s)
+
+                        tareget_basename = f"wellnote_{year_i:04}-{month_i:02}-{day_i:02}_{idx:03}"
+                        target_dir = os.path.join(download_dir, "wellnote", "album", f"{year_i:04}")
+                        target_filepath_woe = os.path.join(target_dir, tareget_basename)
+
+                        # if os.path.exists(target_filepath):
+                        if glob.glob(target_filepath_woe + ".*"):
+                            _LOGGER.warning("Skipping    %s because it exists", target_filepath_woe.replace(os.getcwd(), "."))
+                        else:
+
+                            _LOGGER.debug("Waiting until a clickable vdots button is available")
+                            vdots_button: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "sc-hCwLRM"]))
+                            _LOGGER.info("Clicking vdots button")
+                            vdots_button.click()
+
+                            _LOGGER.debug("Waiting until a clickable download button is available")
+                            download_button: WebElement = wait.until(EC.element_to_be_clickable([By.CLASS_NAME, "sc-hLVXRe"]))
+
+                            with safe_download(driver, wait, download_dir) as download_result:
+                                _LOGGER.warning("Downloading %s", target_filepath_woe.replace(os.getcwd(), "."))
+                                _LOGGER.info("Clicking download button")
+                                download_button.click()
+                                time.sleep(INTERVAL)
+
+                            downloaded_filepath: str = download_result.downloaded_filepath
+
+                            extension: str = downloaded_filepath.split(".")[-1]
+                            target_filepath = target_filepath_woe + "." + extension
+
+                            os.makedirs(os.path.join(target_dir), exist_ok=True)
+                            shutil.move(downloaded_filepath, target_filepath)
+
+                        swiper_button_next = None
+                        with inspect_mode(driver, timeout_sec) as wait2:
+                            _LOGGER.debug("Waiting until a clickable next button is available")
+                            swiper_button_next: WebElement = wait2.until(EC.element_to_be_clickable([By.CLASS_NAME, "swiper-button-next"]))
+                        if not swiper_button_next \
+                        or "swiper-button-disabled" in swiper_button_next.get_attribute("class"):
+                            _LOGGER.info("Breaking this month because next button is not found")
+                            break
+
+                        _LOGGER.info("Clicking the swiper_button_next")
+                        swiper_button_next.click()
+                        # time.sleep(INTERVAL)
+
+                        idx += 1
+
+                    _LOGGER.debug("Waiting until a clickable html body is available")
+                    close_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-edERGn jYcwJ']"]))
+
+                    _LOGGER.info("Closing the preview window")
+                    close_button.click()
+
+                move_next_year_button = None
+                with inspect_mode(driver, timeout_sec) as wait2:
+                    _LOGGER.debug("Waiting until a clickable next year button is available")
+                    move_next_year_button: WebElement = wait2.until(EC.element_to_be_clickable([By.XPATH, "//*[name()='svg' and @class='sc-jFkwbb lbXnTj']"]))
+                if not move_next_year_button:
+                    _LOGGER.info("Breaking this year because next year button is not found")
+                    break
+
+                _LOGGER.info("Moving the next year of %s ", year)
+                move_next_year_button.click()
+                time.sleep(INTERVAL)
+
+                year += 1
+                start_month = 1
     return 0
 
 
